from colusa.etr import Extractor, register_extractor


@register_extractor('//increment.com')
class IncrementDotComExtractor(Extractor):
    def _parse_title(self):
        title = self.bs.find('h1', class_='t-TitleSerif large title')
        return title.text

    def _find_main_content(self):
        return self.bs.find('article', attrs={'itemprop': 'articleBody'})

    def _parse_extra_metadata(self):
        author_tag = self.bs.find('a', attrs={'itemprop': 'author'})
        if author_tag is None:
            author = ''
        else:
            author = author_tag.text

        description = self.bs.find('div', attrs={'itemprop': 'description'})
        if description is None:
            return ''

        return f'author: **{author}**\n\n{description.text}\n'



from bs4 import Tag

from colusa.etr import Extractor, register_extractor_v2
from colusa import logs

@register_extractor_v2('newsweek', '//www.newsweek.com')
class NewsWeekExtractor(Extractor):
    def _find_main_content(self) -> Tag:
        content = self.bs.find('main')
        # structure is: <main> main
        # <article> content
        #  <article>  postContent...
        content = content.find('article') if content else None
        content = content.find('article') if content else None
        return content

    def cleanup(self):
        super().cleanup()


import os
import re
from typing import Any, Callable, Optional, Type, Union

from bs4 import Tag, BeautifulSoup
from dateutil.parser import parse

from .asciidoc_visitor import AsciidocVisitor
from .visitor import NodeVisitor
from .utils import slugify
from .config import BookConfig, MakeConfig, SiteRule

"""
Dictionary of extractor
"""
__EXTRACTORS: dict[str, dict[str, Any]] = {}

"""
Dictionary of transformer
"""
__TRANSFORMERS: dict[str, dict[str, Any]] = {}

"""
Dictionary of postprocessing
"""
__POSTPROCESSORS: dict[str, Type['PostProcessor']] = {}


def register_extractor(pattern: str) -> Callable[[Type['Extractor']], Type['Extractor']]:
    def decorator(cls: Type['Extractor']) -> Type['Extractor']:
        __EXTRACTORS[cls.__name__] = {
            'pattern': pattern,
            'cls': cls,
        }
        return cls

    return decorator


def register_extractor_v2(id: str, pattern: str) -> Callable[[Type['Extractor']], Type['Extractor']]:
    def decorator(cls: Type['Extractor']) -> Type['Extractor']:
        __EXTRACTORS[id] = {
            'pattern': pattern,
            'cls': cls,
        }
        return cls

    return decorator


def register_transformer(pattern: str) -> Callable[[Type['Transformer']], Type['Transformer']]:
    def decorator(cls: Type['Transformer']) -> Type['Transformer']:
        __TRANSFORMERS[cls.__name__] = {
            'pattern': pattern,
            'cls': cls,
        }
        return cls

    return decorator


def register_transformer_v2(id: str, pattern: str) -> Callable[[Type['Transformer']], Type['Transformer']]:
    def decorator(cls: Type['Transformer']) -> Type['Transformer']:
        __TRANSFORMERS[id] = {
            'pattern': pattern,
            'cls': cls,
        }
        return cls

    return decorator


def register_postprocessor(name: str) -> Callable[[Type['PostProcessor']], Type['PostProcessor']]:
    """Register post processing class"""
    def decorator(cls: Type['PostProcessor']) -> Type['PostProcessor']:
        __POSTPROCESSORS[name] = cls
        return cls

    return decorator


def create_extractor(bs: BeautifulSoup, url_path: str, cached_path: str) -> 'Extractor':
    extractor: Extractor = None
    for _, ext in __EXTRACTORS.items():
        p: str = ext['pattern']
        cls: Type['Extractor'] = ext['cls']
        if re.search(p, url_path):
            extractor = cls(bs)
            break
    if extractor is None:
        extractor = Extractor(bs)

    extractor.url_path = url_path
    extractor.cached_path = cached_path
    return extractor

def populate_extractor_config(config: dict[str, Any]) -> None:
    """
    Populate extract configs from external configuration file.
    Should be called at begining before creating any extractor
    """
    for id, v in config.items():
        if id in __EXTRACTORS:
            ext = __EXTRACTORS[id]
            ext.update(v)


def populate_transformer_config(config: dict[str, Any]) -> None:
    """
    Populate transformer configs from external configuration file.
    Should be called at begining before creating any transformer
    """
    for id, v in config.items():
        if id in __TRANSFORMERS:
            trf = __TRANSFORMERS[id]
            trf.update(v)


def create_transformer(url_path: str, content: Tag, root: str) -> 'Transformer':
    config: dict[str, str] = {
        "src_url": url_path,
        "output_dir": root
    }
    for _, trf in __TRANSFORMERS.items():
        p: str = trf['pattern']
        cls: Type['Transformer'] = trf['cls']
        if re.search(p, url_path):
            return cls(config, content)

    return Transformer(config, content)


class ContentNotFoundError(Exception):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        pass

    def __str__(self) -> str:
        return 'cannot detect content of the site'


class Extractor:
    """Extractor extract real article content from sea of other contents"""
    def __init__(self, bs: BeautifulSoup) -> None:
        self.bs: BeautifulSoup = bs
        self.url_path: str = None
        self.cached_path: str = None
        self.content: Optional[Tag] = None
        self.author: Optional[str] = None
        self.published: Optional[str] = None
        self.title: Optional[str] = None
        self.extra_metadata: str = ''

        self.main_content: Optional[Tag] = None

    def parse(self) -> None:
        """
        parse the main content of website and set to self.main_content,
        also parse metadata such as title, author, published date
        """
        self.main_content = self._find_main_content()
        if self.main_content is None:
            raise ContentNotFoundError()

        self._parse_metadata()

    def _parse_title(self) -> str:
        """
        Parse known tags in html for article's title
        :return: title if found
        """
        meta = self.bs.find('meta', attrs={'property': 'og:title'})
        if meta is not None:
            return meta.get('content')
        else:
            return self.bs.title.text

    def _parse_published(self) -> str:
        """
        Parse known tags in html for article's published date
        :return: str value for date in format `%Y-%m-%d`
        """
        meta = self.bs.find('meta', attrs={'property': 'article:published_time'})
        value = meta.get('content') if meta else None
        if value is not None:
            published = parse(value)
            return str(published.date())

        time_published = self.bs.find('time', attrs={'class': 'entry-date published'})
        if time_published is not None:
            return time_published.text

        return ''

    def _parse_author(self) -> str:
        """
        Parse known tags in html for article's author
        :return: author name if found
        """
        meta = self.bs.find('meta', attrs={'name': 'author'})
        if meta is not None:
            value = meta.get('content')
            if value is not None:
                return value
        return ''

    def _parse_extra_metadata(self) -> str:
        return ''

    @classmethod
    def remove_tag(cls, site: Optional[Tag], tag: str, attrs: dict[str, Any]) -> None:
        if site is None:
            return
        elements = site.find_all(tag, attrs=attrs)
        if elements is None:
            return
        for e in elements:
            e.decompose()

    def cleanup(self) -> None:
        """
        Cleanup extra content (mostly ads) within main content
        """
        self.remove_tag(self.main_content, 'div', attrs={'class': 'site-branding'})
        self.remove_tag(self.main_content, 'div', attrs={'class': 'navigation-top'})
        self.remove_tag(self.main_content, 'footer', attrs={})
        # self.remove_tag(self.site, 'footer', attrs={'class': 'site-footer'})
        self.remove_tag(self.main_content, 'div', attrs={'class': 'searchsettings'})
        self.remove_tag(self.main_content, 'section', attrs={'id': 'ajaxsearchlitewidget-2'})
        self.remove_tag(self.main_content, 'aside', attrs={'id': 'secondary'})
        self.remove_tag(self.main_content, 'nav', attrs={'class': 'post-navigation'})
        self.remove_tag(self.main_content, 'header', attrs={'id': 'masthead'})

    def get_content(self) -> Optional[Tag]:
        """
        :return: handle to main content of website
        """
        return self.main_content

    def _find_main_content(self) -> Optional[Tag]:
        """
        The purpose of internal_init is to find the main content of website
        and set the member self.site to handle of that content

        Default implementation tries to cover as much as possible the commonly
        known web structure such as blog, hentry, article

        :return: Tag of main content
        """
        def is_content_class(css_class: Optional[str]) -> bool:
            return css_class is not None and css_class in [
                'postcontent',
                'entry-content',
                'article-content',
                'blog-content',
            ]
        # h-entry from microformat
        site = self.bs.find(class_='hentry')
        if site is not None:
            role_main = site.find('div', attrs={'role': 'main'})
            if role_main is not None:
                site = role_main
            role_main = site.find('div', attrs={'class': 'td-post-content'})
            if role_main is not None:
                site = role_main

            return site

        tag = self.bs.find('div', class_=is_content_class)
        if tag is not None:
            return tag
        hs_blog_post = self.bs.find(attrs={'class': 'hs-blog-post'})
        if hs_blog_post is not None:
            blog_content = hs_blog_post.find(attrs={'class': 'post-body'})
            return blog_content
        tag = self.bs.find('article')
        if tag is not None:
            return tag
        tag = self.bs.find('main')
        if tag is not None:
            return tag
        return None

    def _parse_yoast_data(self) -> dict[str, str]:
        """
        Parse yoast json data to get some metadata such as author, published date
        :return: dict of metadata found in yoast data
        """
        yoast_data = self.bs.find('script', attrs={
            'type': "application/ld+json",
            'class': "yoast-schema-graph",
        })

        return_data: dict[str, str] = {}
        if yoast_data is None:
            return return_data

        import json
        from dateutil import parser
        data = json.loads(yoast_data.string)
        graph = data.get('@graph', [])
        persons: dict[str, str] = {}
        author: Optional[str] = None
        for g in graph:
            if type(g) is not dict:
                continue
            g_type = g.get('@type', '')
            if g_type == 'Article':
                author = g.get('author', {}).get('@id')
                published_value = g.get('datePublished')
                if published_value:
                    published = parser.parse(published_value)
                    return_data['published'] = str(published.date())
                headline = g.get('headline')
                if headline:
                    return_data['title'] = headline

            if (type(g_type) is list and 'Person' in g_type) or (type(g_type) is str and g_type == 'Person'):
                person_id = g.get('@id', '')
                person_name = g.get('name', '')
                persons[person_id] = person_name
        if author in persons:
            return_data['author'] = persons[author]

        return return_data

    def _parse_metadata(self) -> None:
        """
        Parse existing web metadata to get value for `title`, `author`, `published`
        """
        self.title = self._parse_title()
        self.author = self._parse_author()
        self.published = self._parse_published()
        self.extra_metadata = self._parse_extra_metadata()
        data = self._parse_yoast_data()
        self.title = data.get('title', self.title)
        self.author = data.get('author', self.author)
        self.published = data.get('published', self.published)


class DynamicExtractor(Extractor):
    """Extractor driven by user-supplied CSS selectors from a SiteRule."""
    def __init__(self, bs: BeautifulSoup, rule: SiteRule) -> None:
        super().__init__(bs)
        self._rule = rule

    def _find_main_content(self) -> Optional[Tag]:
        if self._rule.content:
            tag = self.bs.select_one(self._rule.content)
            if tag is not None:
                return tag
        return super()._find_main_content()

    def _parse_title(self) -> str:
        if self._rule.title:
            tag = self.bs.select_one(self._rule.title)
            if tag is not None:
                return tag.get_text(strip=True)
        return super()._parse_title()

    def _parse_author(self) -> str:
        if self._rule.author:
            tag = self.bs.select_one(self._rule.author)
            if tag is not None:
                return tag.get_text(strip=True)
        return super()._parse_author()

    def _parse_published(self) -> str:
        if self._rule.published:
            tag = self.bs.select_one(self._rule.published)
            if tag is not None:
                return tag.get_text(strip=True)
        return super()._parse_published()

    def cleanup(self) -> None:
        super().cleanup()
        if self.main_content is None:
            return
        for selector in self._rule.cleanup:
            for el in self.main_content.select(selector):
                el.decompose()


class Transformer:
    """Transformer transform some html tags into asciidoc syntax"""
    def __init__(self, config: dict[str, str], site: Tag) -> None:
        self.value: str = ''
        self.config: dict[str, str] = config
        self.site: Tag = site

    @classmethod
    def tag_wrapper_pre(cls, tag: Tag, text: str, indent: int) -> str:
        code_pattern = r"^\[code\s+lang=(?P<lang>.*)\](?P<content>.*)\[\/code\]$"
        matches = re.finditer(code_pattern, tag.text, re.MULTILINE | re.DOTALL)
        content: list[str] = []
        for matchNum, match in enumerate(matches, start=1):
            ascii_content = f'''[source, {match.group('lang')}]
----
{match.group('content')}
----
'''
            content.append(ascii_content)

        if len(content) == 0:
            print('PRE ====', text)
            content.append(f'''[listing]
....
{text}
....

''')
        return ''.join(content)

    def create_visitor(self) -> NodeVisitor:
        """
        Create new instance of NodeVisitor, used in transform method
        """
        return AsciidocVisitor()

    def transform(self) -> str:
        visitor = self.create_visitor()
        # print(self.site)
        self.value = visitor.visit(self.site, src_url=self.config['src_url'], output_dir=self.config['output_dir'])
        # print(value)
        # cleanup large whitespace
        self.cleanup_after_visit()
        return self.value

    def cleanup_after_visit(self) -> None:
        """
        Cleanup transformed text
        """
        self.value = re.sub(r'(\n\s*){3,}', '\n\n', self.value)


class Render:
    """
        Render book metadata, structure
    """
    def __init__(self, config: Union[dict[str, Any], BookConfig]) -> None:
        # Support both dict and BookConfig for backward compatibility
        if isinstance(config, BookConfig):
            self._config = config
        else:
            self._config = BookConfig.from_dict(config)
        self.output_dir: str = self._config.output_dir
        self.file_list: list[tuple[str, int]] = []

    def render_book_part(self, title: str, description: str) -> None:
        file_name = f'part_{slugify(title)}.asciidoc'
        self.file_list.append((file_name, 0))
        file_path = os.path.join(self.output_dir, file_name)
        with open(file_path, 'w', encoding='utf-8') as file_out:
            file_out.write(f'= {title}\n\n')
            if len(description) > 0:
                file_out.write(description)
                file_out.write('\n\n')

    def render_chapter(self, extractor: Extractor, content: Transformer, src_url: str, basename: str,
                       metadata: bool = True, title_strip: str = '') -> str:
        file_name = f'{basename}.asciidoc'
        self.file_list.append((file_name, 1))
        file_path = os.path.join(self.output_dir, file_name)
        with open(file_path, 'w', encoding='utf-8') as file_out:
            title = extractor.title or ''
            title = title.replace(title_strip, '')
            file_out.write(f'= {title}\n\n')

            if metadata:
                article_metadata = self.render_metadata(extractor, content, src_url)
                file_out.write(article_metadata)

            file_out.write(content.value)

        return file_path

    def render_metadata(self, extractor: Extractor, content: Transformer, src_url: str) -> str:
        from urllib.parse import urlparse

        author = extractor.author
        data: list[str] = []
        if author:
            author_fmt = f'by **{author}**'
            data.append(author_fmt)
        time_published = extractor.published
        if time_published:
            published_info = f'on {time_published}'
            data.append(published_info)

        domain = urlparse(src_url).netloc
        data.append(f'at _link:{src_url}[{domain}]_')
        first_line = ' '.join(data)
        lines: list[str] = [f"{first_line}\n"]
        extra = extractor.extra_metadata
        if extra:
            lines.append(extra)
        lines.append('\n')
        article_metadata = "\n".join(lines)
        return article_metadata

    def render_asciidoc_passthrough(self, file_name: str, title: str) -> None:
        """Add a pre-written AsciiDoc file to the book without transformation."""
        self.file_list.append((file_name, 1))

    def generate_makefile(self, make_params: Union[dict[str, str], MakeConfig]) -> None:
        # Support both dict and MakeConfig
        if isinstance(make_params, MakeConfig):
            html_params = make_params.html
            epub_params = make_params.epub
            pdf_params = make_params.pdf
        else:
            html_params = make_params.get('html', '')
            epub_params = make_params.get('epub', '')
            pdf_params = make_params.get('pdf', '')
        
        book_file_name = self._config.book_file_name
        target_prefix = book_file_name
        target_prefix = target_prefix.removesuffix('.asciidoc').replace('-', '_')
        target_prefix = '' if target_prefix == 'index' else f'{target_prefix}_'
        template = f'''html:
\tasciidoctor {book_file_name} -d book -b html5 -D output {html_params}
\tcp -r images output/

epub:
\tasciidoctor-epub3 {book_file_name} -d book -D output {epub_params}

pdf:
\tasciidoctor-pdf {book_file_name} -d book -D output {pdf_params}
'''
        file_path = os.path.join(self.output_dir, f'{target_prefix}Makefile')
        with open(file_path, 'w') as out_file:
            out_file.write(template)

    def ebook_generate_master_file(self) -> None:
        """ Generate master index.asciidoc to include all book related information such as
        - book title
        - book author
        - book version
        - etc
        - and include all generated asciidoc files from `urls`
        """
        included_files = '\n\n'.join([f'include::{x[0]}[leveloffset={x[1]}]' for x in self.file_list])
        book_properties = '\n'.join([x.strip() for x in self._config.book_properties])
        content = f'''= {self._config.title}
{self._config.author}
{self._config.version}
:doctype: book
:partnums:
:toc:
:imagesdir: images
:homepage: {self._config.homepage}
{book_properties}

{included_files}
'''
        book_file_name = self._config.book_file_name
        with open(os.path.join(self.output_dir, book_file_name), 'w') as index_file:
            index_file.write(content)


class PostProcessor:
    def __init__(self, file_path: str, params: list[Any]) -> None:
        self.file_path: str = file_path
        self.params: list[Any] = params

    def run(self) -> None:
        pass


class PostProcessorNotFoundError(Exception):
    def __init__(self, name: str, *args: Any, **kwargs: Any) -> None:
        self.name: str = name

    def __str__(self) -> str:
        return f'Post Processor {self.name} is not registered'


def create_postprocessor(name: str, file_path: str, params: list[Any]) -> PostProcessor:
    cls = __POSTPROCESSORS.get(name)
    if not cls:
        raise PostProcessorNotFoundError(name)

    return cls(file_path, params)



from colusa.etr import Extractor, register_extractor


@register_extractor('unintendedconsequenc')
class UnintendedConsequencesExtractor(Extractor):
    def _parse_title(self):
        return self.bs.find('h1').text

    def _find_main_content(self):
        return self.bs.find(id='page')

    def get_content(self):
        return self.main_content.find(attrs={'class': 'entry-content'})



from colusa.etr import Extractor, register_extractor


@register_extractor('//www.linkedin.com')
class LinkedInExtractor(Extractor):
    def cleanup(self):
        self.remove_tag(self.main_content, 'header', attrs={})
        super(LinkedInExtractor, self).cleanup()



from colusa.etr import Extractor, register_extractor


@register_extractor('www.cs.rutgers.edu/~pxk/')
class CSRutgersEduExtractor(Extractor):
    def _find_main_content(self):
        return self.bs.find('div', attrs={'id': 'main'})

    def cleanup(self):
        self.remove_tag(self.main_content, 'div', attrs={'id': 'downloadmsg'})
        self.remove_tag(self.main_content, 'div', attrs={'id': 'headline'})
        super(CSRutgersEduExtractor, self).cleanup()



from bs4 import Tag

from colusa.etr import Extractor, register_extractor_v2
from colusa import logs

@register_extractor_v2('stratechery', '//stratechery.com')
class StratecheryExtractor(Extractor):
    def _find_main_content(self) -> Tag:
        content = self.bs.find('main')
        return content

    def cleanup(self):
        self.remove_tag(self.main_content, 'div', attrs={'class': 'sharedaddy'})
        self.remove_tag(self.main_content, 'nav', attrs={})
        self.remove_tag(self.main_content, 'div', attrs={'class': 'post-navigation-link-previous'})
        self.remove_tag(self.main_content, 'div', attrs={'class': 'post-navigation-link-next'})
        self.remove_tag(self.main_content, 'div', attrs={'class': 'stratechery-sidebar'})
        super().cleanup()


# -*- coding: utf-8 -*-
"""Top-level package for colusa.

Copyright (c) 2020-2024 Huu Hoa NGUYEN

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from colusa import _version

__author__ = _version.__author__
__email__ = _version.__email__
__version__ = _version.__version__
__copyright__ = _version.__copyright__
__license__ = _version.__license__

__all__ = [
    'Colusa',
    'ConfigurationError',
    'logs',
    'Crawler',
    'BookConfig',
    'MakeConfig',
    'PartConfig',
    'PostProcessingConfig',
]


from colusa import logs
from .crawlers import Crawler as Crawler
from colusa import etr as etr
from colusa.exceptions import ConfigurationError as ConfigurationError
from colusa import utils as utils
from colusa.colusa import Colusa as Colusa
from colusa.config import (
    BookConfig as BookConfig,
    MakeConfig as MakeConfig,
    PartConfig as PartConfig,
    PostProcessingConfig as PostProcessingConfig,
)



import re

from colusa.etr import Extractor, register_extractor


@register_extractor('//untools.co')
class UntoolsExtractor(Extractor):
    def _parse_title(self):
        import re
        header = self.bs.find('div', class_=re.compile('article-module--top--'))
        title = header.find('h2').text
        return title

    def _parse_extra_metadata(self):
        import re
        header = self.bs.find('div', class_=re.compile('article-module--top--'))
        tag = header.find('span', class_=re.compile('tag-module--tag--')).text
        usage = header.find('div', class_=re.compile('article-module--when-useful--')).text

        return f'.{tag}\n****\n{usage}\n****\n\n'

    def _find_main_content(self):
        return self.bs.find(class_=re.compile('article-module--content--'))



# -*- coding: utf-8 -*-
"""Configuration dataclasses for type-safe configuration handling."""

from dataclasses import dataclass, field
from typing import Any, Optional


@dataclass
class UrlEntry:
    """A single URL or local file path entry with optional metadata overrides."""
    path: str
    title: Optional[str] = None
    author: Optional[str] = None
    published: Optional[str] = None


def _parse_url_entry(entry: Any) -> 'UrlEntry':
    """Convert a raw config entry (str or dict) into a UrlEntry."""
    if isinstance(entry, str):
        return UrlEntry(path=entry)
    return UrlEntry(
        path=entry['path'],
        title=entry.get('title'),
        author=entry.get('author'),
        published=entry.get('published'),
    )


@dataclass
class MakeConfig:
    """Configuration for makefile generation (html, epub, pdf targets)."""
    html: str = ''
    epub: str = ''
    pdf: str = ''


@dataclass
class PostProcessingConfig:
    """Configuration for a post-processing step."""
    processor: str
    params: list[Any] = field(default_factory=list)


@dataclass
class PartConfig:
    """Configuration for a book part (used in multi-part books)."""
    title: str
    description: str = ''
    urls: list[UrlEntry] = field(default_factory=list)


@dataclass
class DownloaderConfig:
    """Configuration for the content downloader."""
    # Extensible for different fetcher configs (e.g., tor, proxy, etc.)
    pass


@dataclass
class SiteRule:
    """User-defined CSS-selector-based parsing rule for a site pattern."""
    pattern: str
    content: Optional[str] = None
    title: Optional[str] = None
    author: Optional[str] = None
    published: Optional[str] = None
    cleanup: list[str] = field(default_factory=list)


def _parse_site_rule(d: dict[str, Any]) -> SiteRule:
    return SiteRule(
        pattern=d['pattern'],
        content=d.get('content'),
        title=d.get('title'),
        author=d.get('author'),
        published=d.get('published'),
        cleanup=d.get('cleanup', []),
    )


@dataclass
class BookConfig:
    """Main configuration for generating an ebook.
    
    Attributes:
        title: Book title
        author: Book author
        version: Book version string
        homepage: URL to the book's homepage
        output_dir: Directory for generated output
        book_file_name: Name of the main asciidoc file
        multi_part: Whether the book has multiple parts
        metadata: Whether to include article metadata
        make: Makefile generation settings
        postprocessing: List of post-processing steps
        parts: List of book parts (for multi-part books)
        urls: List of URLs to process (for single-part books)
        book_properties: Additional asciidoc book properties
        title_prefix_trim: Prefix to strip from article titles
        downloader: Downloader configuration
        extractors: Extractor-specific configurations
        transformers: Transformer-specific configurations
    """
    title: str
    author: str
    version: str
    homepage: str
    output_dir: str = '.'
    book_file_name: str = 'index.asciidoc'
    multi_part: bool = False
    metadata: bool = True
    make: MakeConfig = field(default_factory=MakeConfig)
    postprocessing: list[PostProcessingConfig] = field(default_factory=list)
    parts: list[PartConfig] = field(default_factory=list)
    urls: list[UrlEntry] = field(default_factory=list)
    book_properties: list[str] = field(default_factory=list)
    title_prefix_trim: str = ''
    downloader: dict[str, Any] = field(default_factory=dict)
    extractors: dict[str, Any] = field(default_factory=dict)
    transformers: dict[str, Any] = field(default_factory=dict)
    site_rules: list[SiteRule] = field(default_factory=list)
    site_rules_file: str = ''

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> 'BookConfig':
        """Create a BookConfig from a dictionary (e.g., loaded from JSON/YAML).
        
        Args:
            data: Dictionary containing configuration data
            
        Returns:
            BookConfig instance with values from the dictionary
        """
        make_data = data.get('make', {})
        make_config = MakeConfig(
            html=make_data.get('html', ''),
            epub=make_data.get('epub', ''),
            pdf=make_data.get('pdf', ''),
        )
        
        postprocessing = [
            PostProcessingConfig(
                processor=pp.get('processor', ''),
                params=pp.get('params', [])
            )
            for pp in data.get('postprocessing', [])
        ]
        
        parts = [
            PartConfig(
                title=part.get('title', ''),
                description=part.get('description', ''),
                urls=[_parse_url_entry(e) for e in part.get('urls', [])]
            )
            for part in data.get('parts', [])
        ]

        return cls(
            title=data.get('title', ''),
            author=data.get('author', ''),
            version=data.get('version', ''),
            homepage=data.get('homepage', ''),
            output_dir=data.get('output_dir', '.'),
            book_file_name=data.get('book_file_name', 'index.asciidoc'),
            multi_part=data.get('multi_part', False),
            metadata=data.get('metadata', True),
            make=make_config,
            postprocessing=postprocessing,
            parts=parts,
            urls=[_parse_url_entry(e) for e in data.get('urls', [])],
            book_properties=data.get('book_properties', []),
            title_prefix_trim=data.get('title_prefix_trim', ''),
            downloader=data.get('downloader', {}),
            extractors=data.get('extractors', {}),
            transformers=data.get('transformers', {}),
            site_rules=[_parse_site_rule(r) for r in data.get('site_rules', [])],
            site_rules_file=data.get('site_rules_file', ''),
        )

    def to_dict(self) -> dict[str, Any]:
        """Convert the BookConfig back to a dictionary.
        
        Returns:
            Dictionary representation of the configuration
        """
        return {
            'title': self.title,
            'author': self.author,
            'version': self.version,
            'homepage': self.homepage,
            'output_dir': self.output_dir,
            'book_file_name': self.book_file_name,
            'multi_part': self.multi_part,
            'metadata': self.metadata,
            'make': {
                'html': self.make.html,
                'epub': self.make.epub,
                'pdf': self.make.pdf,
            },
            'postprocessing': [
                {'processor': pp.processor, 'params': pp.params}
                for pp in self.postprocessing
            ],
            'parts': [
                {'title': p.title, 'description': p.description, 'urls': p.urls}
                for p in self.parts
            ],
            'urls': self.urls,
            'book_properties': self.book_properties,
            'title_prefix_trim': self.title_prefix_trim,
            'downloader': self.downloader,
            'extractors': self.extractors,
            'transformers': self.transformers,
        }



from colusa.etr import Extractor, register_extractor


@register_extractor('//www.infoq.com')
class InfoQExtractor(Extractor):
    def _find_main_content(self):
        return self.bs.find('div', class_='article__content')

    def cleanup(self):
        self.remove_tag(self.main_content, 'div', attrs={'class': 'contentRatingWidget'})
        self.remove_tag(self.main_content, 'div', attrs={'class': 'widget article__fromTopic topics'})
        self.remove_tag(self.main_content, 'div', attrs={'class': 'nocontent'})
        super(InfoQExtractor, self).cleanup()



from colusa.etr import Extractor, register_extractor


@register_extractor('.wikipedia.org/')
class WikipediaExtractor(Extractor):
    def _parse_title(self):
        title = self.bs.find('h1', id='firstHeading')
        return title.text

    def _find_main_content(self):
        return self.bs.find(id='bodyContent')



from colusa.etr import Extractor, register_extractor


@register_extractor('//scrumcrazy.wordpress.com')
class ScrumCrazyExtractor(Extractor):
    def __init__(self, bs):
        self._cached_author = None
        super(ScrumCrazyExtractor, self).__init__(bs)

    def _parse_author(self) -> str:
        if self._cached_author:
            return self._cached_author
        return super(ScrumCrazyExtractor, self)._parse_author()

    def _find_main_content(self):
        possible_main = super(ScrumCrazyExtractor, self)._find_main_content()
        post_info_div = possible_main.find('div', class_='postinfo')
        if post_info_div is not None:
            post_info = post_info_div.text.strip()
            import re
            m = re.search(r'by (.*)$', post_info)
            if m is not None:
                self._cached_author = m.group(1)
        entry = possible_main.find('div', class_='entry')
        if entry is not None:
            return entry
        else:
            return possible_main

    def cleanup(self):
        self.remove_tag(self.main_content, 'div', attrs={'id': 'jp-post-flair'})
        self.remove_tag(self.main_content, 'p', attrs={'class': 'postinfo'})



from typing import Any, TextIO
from colusa import logs
from colusa.fetch import Downloader
from bs4 import BeautifulSoup
import pathlib
from collections import OrderedDict
from urllib.parse import urljoin
from colusa import utils


class Crawler:
    def __init__(self, url: str, output_dir: str, output_file: TextIO) -> None:
        self.url: str = url
        self.output_dir: str = output_dir
        self.output_file: TextIO = output_file
        self.downloader: Downloader = Downloader()
        self.content: str = ''
        self.dom: BeautifulSoup

    def run(self) -> None:
        logs.info(f"hello: {self.url}")
        self.content = self.download_content()
        self.dom = BeautifulSoup(self.content, 'html.parser')

        table_chapter = self.dom.find('table', id='chapters')
        anchors = table_chapter.find_all('a')
        urls: OrderedDict[str, bool] = OrderedDict()
        for a in anchors:
            href = a.attrs.get('href')
            final_url = urljoin(self.url, href)
            urls[final_url] = True
        
        import json
        result: list[str] = [str(k) for k in urls.keys()]
        json.dump(result, self.output_file, indent=3)

    def download_content(self) -> str:
        """
        Download html content of given `url_path` then cached in `.cached` folder of local file system
        :param url_path: url of html article
        :return: content of downloaded file
        """
        output_path = pathlib.Path(self.output_dir)
        cached_file_path = output_path.joinpath('.cached', f'{utils.get_hexdigest(self.url)}.html')
        logs.info(self.url, cached_file_path)

        if not cached_file_path.exists():
            # download file from url_path
            self.downloader.download_url(self.url, str(cached_file_path))

        with open(cached_file_path, 'rt', encoding='utf-8') as file_in:
            content = file_in.read()
        return content



from typing import Any, Optional, Union
from bs4 import BeautifulSoup
import pathlib
import json
import yaml
import os

from colusa import logs, etr, utils, fetch, ConfigurationError
from colusa.config import BookConfig, MakeConfig, UrlEntry, SiteRule, _parse_site_rule


def _load_rules_file(path: str) -> list[SiteRule]:
    p = pathlib.PurePath(path)
    with open(path, 'r', encoding='utf-8') as f:
        data = yaml.safe_load(f) if p.suffix == '.yml' else json.load(f)
    return [_parse_site_rule(r) for r in (data or [])]


class Colusa:
    """
    Implementation for initializing book configuration and generating book from configuration
    """
    def __init__(self, configuration: Union[dict[str, Any], BookConfig], config_file_dir: str = '') -> None:
        from colusa.etr import populate_extractor_config, populate_transformer_config

        utils.scan('colusa.plugins')

        # Support both dict and BookConfig for backward compatibility
        if isinstance(configuration, BookConfig):
            self.config = configuration
        else:
            self.config = BookConfig.from_dict(configuration)

        self.output_dir = self.config.output_dir
        self.book_maker = etr.Render(self.config)
        self.downloader = fetch.Downloader(self.config.downloader)
        populate_extractor_config(self.config.extractors)
        populate_transformer_config(self.config.transformers)
        self.site_rules: list[SiteRule] = self._load_site_rules(config_file_dir)

    @classmethod
    def generate_new_configuration(cls, file_path: str) -> None:
        """
        Generate initial content for book configuration, raise ConfigurationError if
        `file_path` is not end with supported format (json, yml)

        :param file_path: path of new configuration file, must end with either .json or .yml
        """
        template = {
            "title": "__fill the title__",
            "author": "__fill the author__",
            "version": "v1.0",
            "homepage": "__fill url to home page__",
            "output_dir": "__fill output dir__",
            "book_file_name": "index.asciidoc",
            "multi_part": False,
            "metadata": True,
            "make": {
                'html': '',
                'epub': '',
                'pdf': '',
            },
            'postprocessing': [],
            "parts": [],
            "urls": []
        }
        p = pathlib.PurePath(file_path)
        if p.suffix == '.json':
            with open(file_path, 'w') as file_out:
                json.dump(template, file_out, indent=4)
                return
        if p.suffix == '.yml':
            with open(file_path, 'w') as file_in:
                yaml.safe_dump(template, file_in)
                return
        raise ConfigurationError(f'unknown configuration file format: {p.suffix}. '
                                 f'Configuration file format should be either .json or .yml')

    def _load_site_rules(self, config_file_dir: str) -> list[SiteRule]:
        rules = list(self.config.site_rules)
        rules_file = self.config.site_rules_file
        if rules_file:
            if not os.path.isabs(rules_file) and config_file_dir:
                rules_file = os.path.join(config_file_dir, rules_file)
            rules.extend(_load_rules_file(rules_file))
        return rules

    def _match_site_rule(self, url_path: str) -> Optional[SiteRule]:
        import re
        for rule in self.site_rules:
            if re.search(rule.pattern, url_path):
                return rule
        return None

    @classmethod
    def generate_book(cls, config_file_path: str) -> None:
        """
        Generate book from existing input configuration `config_file_path`,
        raise ConfigurationError if cannot parse configuration file

        :param config_file_path: Path to input configuration
        :return: None
        """
        configs = cls._read_configuration_file(config_file_path)
        config_dir = str(pathlib.Path(config_file_path).parent)
        with Colusa(configs, config_file_dir=config_dir) as s:
            s.generate()

    @classmethod
    def _read_configuration_file(cls, file_path: str) -> BookConfig:
        """
        Read the book configuration file, raise ConfigurationError if `file_path` is unknown format

        :param file_path: path to known format configuration (json, yml)
        :return: BookConfig instance
        """
        p = pathlib.PurePath(file_path)
        if p.suffix == '.json':
            with open(file_path, 'r') as file_in:
                data = json.load(file_in)
                return BookConfig.from_dict(data)
        if p.suffix == '.yml':
            with open(file_path, 'r') as file_in:
                data = yaml.safe_load(file_in)
                return BookConfig.from_dict(data)
        raise ConfigurationError(f'unknown configuration file format: {p.suffix}. '
                                 f'Configuration file format should be either .json or .yml')

    @staticmethod
    def _is_local_path(path: str) -> bool:
        return not path.startswith(('http://', 'https://'))

    def download_content(self, url_entry: UrlEntry) -> str:
        """
        Return html content for the given entry.
        For local paths, reads the file directly.
        For remote URLs, downloads and caches as before.
        :param url_entry: UrlEntry with path and optional metadata
        :return: content of the file
        """
        url_path = url_entry.path

        if self._is_local_path(url_path):
            import chardet
            with open(url_path, 'rb') as f:
                result = chardet.detect(f.read())
                encoding = result['encoding']
            with open(url_path, 'rt', encoding=encoding) as file_in:
                return file_in.read()

        import chardet

        output_path = pathlib.Path(self.output_dir)
        cached_file_path = output_path.joinpath('.cached', f'{utils.get_hexdigest(url_path)}.html')
        logs.info(url_path, cached_file_path)

        if not cached_file_path.exists():
            # download file from url_path
            self.downloader.download_url(url_path, str(cached_file_path))

        with open(cached_file_path, 'rb') as f:
            result = chardet.detect(f.read())
            encoding = result['encoding']
        with open(cached_file_path, 'rt', encoding=encoding) as file_in:
            content = file_in.read()
        return content

    @staticmethod
    def _get_saved_file_name(url_path: str) -> str:
        """
        calculates the name on local file system based on given article url.
        The calculation is based on ending name of url and url's short digest

        saved_file_name =  f'{pathlib.PurePath(url_path).name}_{get_short_hexdigest(url_path)}'

        Args:
            url_path (str): url

        Returns:
            str: calculated file name
        """
        p = pathlib.PurePath(url_path)
        return f'{p.name}_{utils.get_short_hexdigest(url_path)}'

    def __enter__(self) -> 'Colusa':
        return self

    def __exit__(self, *args: Any) -> None:
        self.close()
    
    def close(self) -> None:
        self.downloader.close()

    def _process_local_asciidoc(self, url_entry: UrlEntry) -> None:
        """Copy a local .adoc file into the output dir and register it in the book."""
        import shutil
        src_path = pathlib.Path(url_entry.path)
        file_name = self._get_saved_file_name(url_entry.path) + '.asciidoc'
        dest_path = pathlib.Path(self.output_dir) / file_name
        shutil.copy2(src_path, dest_path)

        if url_entry.title:
            title = url_entry.title
        else:
            # Try to parse from first `= ...` line
            title = src_path.stem
            try:
                with open(src_path, 'r', encoding='utf-8') as f:
                    for line in f:
                        line = line.strip()
                        if line.startswith('= '):
                            title = line[2:].strip()
                            break
            except Exception:
                pass

        self.book_maker.render_asciidoc_passthrough(file_name, title)

    def ebook_generate_content(self, url_entry: UrlEntry) -> None:
        url_path = url_entry.path

        # Route local .adoc files directly — no HTML extraction needed
        if self._is_local_path(url_path) and pathlib.PurePath(url_path).suffix in ('.adoc', '.asciidoc'):
            self._process_local_asciidoc(url_entry)
            return

        content = self.download_content(url_entry)
        bs = BeautifulSoup(content, 'html.parser')

        chapter_metadata = self.config.metadata
        title_strip = self.config.title_prefix_trim
        try:
            rule = self._match_site_rule(url_path)
            if rule:
                extractor = etr.DynamicExtractor(bs, rule)
                extractor.url_path = url_path
                extractor.cached_path = os.path.join(self.output_dir, '.cached')
            else:
                extractor = etr.create_extractor(bs, url_path, os.path.join(self.output_dir, '.cached'))
            extractor.parse()
            extractor.cleanup()
            # Apply per-entry metadata overrides
            if url_entry.title:
                extractor.title = url_entry.title
            if url_entry.author:
                extractor.author = url_entry.author
            if url_entry.published:
                extractor.published = url_entry.published
            transformer = etr.create_transformer(url_path, extractor.get_content(), self.output_dir)
            transformer.transform()
            file_path = self.book_maker.render_chapter(extractor, transformer, url_path,
                                           self._get_saved_file_name(url_path),
                                           metadata=chapter_metadata,
                                           title_strip=title_strip)
            for pp in self.config.postprocessing:
                pp_name = pp.processor
                pp_params = pp.params
                pcls = etr.create_postprocessor(pp_name, file_path, pp_params)
                pcls.run()
        except etr.ContentNotFoundError as e:
            logs.error(e, url_path)
            # raise e

    def generate(self) -> None:
        os.makedirs(os.path.join(self.output_dir, ".cached"), exist_ok=True)
        os.makedirs(os.path.join(self.output_dir, "images"), exist_ok=True)
        self.book_maker.generate_makefile(self.config.make)

        if self.config.multi_part:
            self._generate_book_multi_part()
        else:
            self._generate_book_single_part()

        self.book_maker.ebook_generate_master_file()

    def _generate_book_single_part(self) -> None:
        entries = self.config.urls
        if len(entries) == 0:
            raise ConfigurationError('urls field must contain at least one url')
        for entry in entries:
            self.ebook_generate_content(entry)

    def _generate_book_multi_part(self) -> None:
        parts = self.config.parts
        if len(parts) == 0:
            raise ConfigurationError('parts field must contain at least one part object')
        for part in parts:
            self.book_maker.render_book_part(part.title, part.description)
            for entry in part.urls:
                self.ebook_generate_content(entry)



from colusa.etr import Extractor, register_extractor


@register_extractor('//xp123.com')
class XP123Extractor(Extractor):
    def _find_main_content(self):
        return self.bs.find('article', attrs={'class': 'post'})

    def cleanup(self):
        self.remove_tag(self.main_content, 'header', attrs={'class': 'entry-header'})
        self.remove_tag(self.main_content, 'section', attrs={'class': 'yikes-mailchimp-container'})
        self.remove_tag(self.main_content, 'footer', attrs={'class': 'entry-meta'})



from colusa.etr import Extractor, register_extractor


@register_extractor('//www.preethikasireddy.com')
class PreethikasireddyExtractor(Extractor):
    def _find_main_content(self):
        return self.bs.find('div', attrs={'id': 'Story'})



from colusa.etr import Extractor, register_extractor


@register_extractor('//knowledgegraph.today')
class KnowledgegraphTodayExtractor(Extractor):
    def _find_main_content(self):
        return self.bs.body

    def _parse_title(self):
        # TODO: self.main_content here might not be available, need to find other way
        p_title = self.main_content.find('p', class_='title')
        return p_title.text



import unittest

from colusa import Colusa
from colusa.config import UrlEntry
from colusa.utils import get_short_hexdigest, get_hexdigest
import pathlib
import os
from unittest.mock import patch
import os.path


def download_image_mocked(url_path, *args, **kwargs):
    import urllib
    print(f'call download_image with {url_path}')
    result = urllib.parse.urlsplit(url_path)
    p = pathlib.PurePath(result.path)
    image_name = f'{get_hexdigest(url_path)}{p.suffix}'
    return image_name


def download_content_mocked(url_entry: UrlEntry):
    url_path = url_entry.path
    print(f'call colusa.Colusa.download_content with path {url_path}')
    cached_file_path = os.path.join('tests-cached', f'{get_hexdigest(url_path)}.html')
    with open(cached_file_path, 'rt', encoding='utf-8') as file_in:
        content = file_in.read()
    return content


class ColusaIntegrationTestCase(unittest.TestCase):

    @patch('os.path.exists')
    def test_patch_exists(self, mock_os_path_exists):
        def side_effect(path: str):
            print(f'call os.path.exists with {path}')
            return path.startswith('tests-dist/images/')

        mock_os_path_exists.side_effect = side_effect

        self.assertTrue(os.path.exists('tests-dist/images/abc.png'))
        self.assertFalse(os.path.exists('tests-dist/.cached/images/abc.png'))

    @patch('colusa.Colusa.download_content')
    def test_mock_download_content(self, mock_download_content):
        def side_effect(url_entry: UrlEntry):
            url_path = url_entry.path
            print(f'call colusa.Colusa.download_content with path {url_path}')
            cached_file_path = os.path.join('tests-cached', f'{get_hexdigest(url_path)}.html')
            with open(cached_file_path, 'rt', encoding='utf-8') as file_in:
                content = file_in.read()
            return content

        mock_download_content.side_effect = side_effect

        configs = {
            "title": "test: test case",
            "author": "tester",
            "version": "v1.0",
            "homepage": "dummy",
            "output_dir": "tests-dist",
            "multi_part": False,
            "parts": [],
            "urls": [
                "https://staffeng.com/guides/overview-overview",
            ]
        }
        runner = Colusa(configs)
        runner.generate()
        mock_download_content.assert_called_once_with(UrlEntry(path="https://staffeng.com/guides/overview-overview"))

    @patch('colusa.Colusa.download_content')
    @patch('colusa.fetch.download_image')
    def test_whole_program(self, mock_download_image, mock_download_content):
        urls = [
            "https://staffeng.com/guides/overview-overview",
            "https://tech.trivago.com/2016/03/24/team-work-made-simple-with-guilds/",
            "https://techblog.constantcontact.com/software-development/tackling-tech-debt-frontend-guild/",
            "https://medium.com/@shawnstafford/ops-runbook-16017fa78733",
            "https://martinfowler.com/articles/patterns-legacy-displacement/",
            "https://doordash.engineering/2021/03/04/building-a-declarative-real-time-feature-engineering-framework/",
        ]

        """
        Mocking os.path.exists with using side_effect
        """
        mock_download_image.side_effect = download_image_mocked
        mock_download_content.side_effect = download_content_mocked

        for url in urls:
            configs = {
                "title": "test: test case",
                "author": "tester",
                "version": "v1.0",
                "homepage": "dummy",
                "output_dir": "tests-dist",
                "multi_part": False,
                "parts": [],
                "urls": [
                    url
                ]
            }
            runner = Colusa(configs)
            runner.generate()

            p = pathlib.PurePath(url)
            output_name = f'{p.name}_{get_short_hexdigest(url)}.asciidoc'
            self.assertTrue(self.compare_file(os.path.join('tests-dist', output_name),
                                              os.path.join('tests-expected', output_name)),
                            f'failed for url: {url}')

    def compare_file(self, actual_file, expected_file):
        import filecmp
        return filecmp.cmp(actual_file, expected_file)


if __name__ == '__main__':
    unittest.main()



import argparse
import sys
from typing import Any

from colusa import Colusa, ConfigurationError, logs


def main() -> None:
    args = parse_args()
    args.func(args)


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()

    commands = parser.add_subparsers(dest='command')
    commands.required = True

    init_parser = commands.add_parser('init',
                                      help='Generate configuration file with basic structure which '
                                           'contains every input fields required by colusa')
    init_parser.set_defaults(func=init)
    init_parser.add_argument('output', type=str, help='Configuration file. '
                                                      'File extension should be either json or yml')

    generate_parser = commands.add_parser('generate',
                                          help='Generate ebook source data based on input configuration')
    generate_parser.set_defaults(func=generate)
    generate_parser.add_argument('input', type=str, help='Configuration file. '
                                                         'File extension should be either json or yml')

    crawler_parse = commands.add_parser('crawl',
                        help='Crawl an URL to generate list of URLs')
    crawler_parse.set_defaults(func=crawl_url)
    crawler_parse.add_argument('--url', '-u', type=str, help='URL to crawl')
    crawler_parse.add_argument('--output_dir', '-d', type=str, help='Output folder to store cached')
    crawler_parse.add_argument('--output', '-o', type=argparse.FileType('w'), default=sys.stdout, help='Output file (default: standard output)')

    return parser.parse_args()


def init(args: argparse.Namespace) -> None:
    try:
        Colusa.generate_new_configuration(args.output)
    except ConfigurationError as e:
        logs.error(e)


def generate(args: argparse.Namespace) -> None:
    try:
        Colusa.generate_book(args.input)
    except ConfigurationError as e:
        logs.error(e)


def crawl_url(args: argparse.Namespace) -> None:
    from colusa import Crawler
    try:
        crawler = Crawler(args.url, args.output_dir, args.output)
        crawler.run()
    except ConfigurationError as e:
        logs.error(e)


if __name__ == '__main__':
    main()



from colusa.etr import Extractor, register_extractor


@register_extractor('//staffeng.com')
class StaffEng(Extractor):
    def _find_main_content(self):
        return self.bs.find(class_='blog-post-content')



from bs4 import Tag

from colusa.etr import Extractor, register_extractor_v2
from colusa import logs

@register_extractor_v2('sethsblog', '//seths.blog')
class SethsBlogExtractor(Extractor):
    def _find_main_content(self) -> Tag:
        content = self.bs.find('div', class_='single-post')
        return content

    def cleanup(self):
        self.remove_tag(self.main_content, 'div', attrs={'class': 'post-footer'})
        super().cleanup()


from colusa.etr import Extractor, register_extractor


@register_extractor('avikdas.com')
class AvikdasExtractor(Extractor):
    def _find_main_content(self):
        return self.bs.body

    def cleanup(self):
        self.remove_tag(self.main_content, 'ul', attrs={'id': 'bottom-links'})
        self.remove_tag(self.main_content, 'h1', attrs={})



import re

from bs4 import Tag

from colusa.etr import Extractor, Transformer, register_extractor, register_transformer


@register_extractor('//metruyenchu.com')
class MeTruyenChuExtractor(Extractor):
    def _find_main_content(self):
        return self.bs.find('div', id='js-read__content')

    def cleanup(self):
        self.remove_tag(self.main_content, 'div', attrs={'id': 'js-left-menu'})
        self.remove_tag(self.main_content, 'div', attrs={'id': 'js-right-menu'})
        to_remove = []
        to_remove += self.main_content.select('#js-read__body div:first-child')
        to_remove += self.main_content.select('#js-read__body div:nth-child(2)')
        to_remove += self.main_content.select('#js-read__body div:nth-child(3)')
        for i in to_remove:
            i.extract()
        
        self.remove_tag(self.main_content, 'a', attrs={'class': 'truyen-title'})
        self.remove_tag(self.main_content, 'h2', attrs={})
        self.remove_tag(self.main_content, 'div', attrs={'class': 'group_story'})
        


@register_transformer('//metruyenchu.com')
class MeTruyenChuTransformer(Transformer):
    def transform(self):
        value = super().transform()
        # cleanup hr
        value = re.sub(r'^-{3,}', "'''", value, flags=re.MULTILINE)
        value = re.sub(r'^\.{4,}', "'''", value, flags=re.MULTILINE)
        value = re.sub(r'^_{3,}', "'''", value, flags=re.MULTILINE)
        value = re.sub(r'^\*{3,}', "'''", value, flags=re.MULTILINE)
        value = re.sub(r'^\+{5,}', "'''", value, flags=re.MULTILINE)
        value = re.sub(r"^'''\n\n$", "", value, flags=re.MULTILINE)
        value = re.sub(r"^'''\n{2,}'''", "", value, flags=re.MULTILINE)
        if value.startswith(' '):
            value = value[1:]
        self.value = value
        return value




from bs4 import Tag

from colusa.etr import Extractor, Transformer, register_extractor, register_transformer


@register_extractor('//medium.com')
class MediumExtractor(Extractor):
    def _parse_title(self):
        title = self.main_content.find('h1')
        if title is not None:
            return title.text
        meta = self.bs.find('meta', attrs={'property': 'og:title'})
        if meta is not None:
            return meta.get('content')
        return self.bs.title.text

    def _find_main_content(self):
        return self.bs.find('article')


@register_transformer('//medium.com')
class MediumTransformer(Transformer):
    def transform_tag(self, tag: Tag, indent_level=0) -> str:
        if tag.name == 'figure' and 'paragraph-image' in tag.get('class'):
            tag = tag.find('noscript')
            value = super(MediumTransformer, self).transform_tag(tag, indent_level)
            return f'{value}\n\n'
        return super(MediumTransformer, self).transform_tag(tag, indent_level)



from colusa.etr import Extractor, register_extractor


@register_extractor('//blog.acolyer.org')
class TheMorningPaperExtractor(Extractor):
    def _parse_title(self):
        header = self.bs.find('h1', class_='entry-title')
        return header.text




from colusa.etr import Extractor, register_extractor


@register_extractor('https://cadenceworkflow.io')
class CadenceWorkflowExtractor(Extractor):
    def _find_main_content(self):
        return self.bs.find('main')



import unittest

from colusa.etr import Transformer
from bs4 import BeautifulSoup


def read_file(file_path):
    with open(file_path) as file_in:
        data = file_in.read()

    return data


class TransformerTestCase(unittest.TestCase):
    def test_tag_p_1(self):
        test_data = read_file('tests/data/p_1.html')
        expected_output = read_file('tests/data/p_1.txt')
        bs = BeautifulSoup(test_data, 'html.parser')
        sp = Transformer({
            'src_url': 'https://dummy',
            'output_dir': 'temp'
        }, bs)
        actual_data = sp.transform()

        self.assertEqual(expected_output, actual_data)


if __name__ == '__main__':
    unittest.main()



from colusa.etr import Extractor, register_extractor


@register_extractor('//eugeneyan.com')
class EugeneyanExtractor(Extractor):
    def _parse_title(self):
        title = self.bs.find('h1', class_='title')
        return title.text

    def _find_main_content(self):
        return self.bs.find('div', class_='notebody')
